// Copyright 2020 Arthur Sonzogni. All rights reserved.
// Use of this source code is governed by the MIT license that can be found in
// the LICENSE file.
#include <stdlib.h>                   // for EXIT_SUCCESS
#include <chrono>                     // for milliseconds
#include <fmt/core.h>
#include <ftxui/component/event.hpp>  // for Event
#include <ftxui/component/mouse.hpp>  // for ftxui
#include <ftxui/dom/elements.hpp>  // for text, separator, Element, operator|, vbox, border
#include <memory>                  // for allocator, shared_ptr
#include <string>                  // for operator+, to_string
#include <thread>                  // for sleep_for

#include "ftxui/component/component.hpp"  // for CatchEvent, Renderer, operator|=
#include "ftxui/component/loop.hpp"       // for Loop
#include "ftxui/component/screen_interactive.hpp"  // for ScreenInteractive
#include <vector>

using namespace ftxui;
using namespace std;

ButtonOption Style() {
  auto option = ButtonOption::Animated();
  option.transform = [](const EntryState& s) {
    auto element = text(s.label);
    if (s.focused) {
      element |= bold;
    }
    return element | center | borderEmpty | flex;
  };
  return option;
}

int main() {
  auto screen = ScreenInteractive::Fullscreen();
  screen.TrackMouse(true);
  vector<string> lines;

  // Create a component counting the number of frames drawn and event handled.
  float scroll_x = 0.1;
  float scroll_y = 1.0;
  int hp = 100;

  auto status = Renderer([&] {
      return vbox({
        paragraph(fmt::format("HP {}", hp)),
        separator(),
        hbox({
            text("HP"),
            gauge(hp / 100.0f),
          }),
      });
  });

  auto content = Renderer([&] {
      vector<Element> output;

      for(const auto line : lines) {
        output.push_back(text(line));
      }

      return vbox(output);
    });

  auto scrollable_content = Renderer(content,
      [&, content] {
      return content->Render()
        | focusPositionRelative(scroll_x, scroll_y)
        | frame | flex;
  });

  auto component = Renderer(scrollable_content, [&] {
    return vbox({
        status->Render(),
        separator(),
        scrollable_content->Render() | vscroll_indicator | yframe | size(HEIGHT, LESS_THAN, 20),
        }) |
    border;
  });

  component |= CatchEvent([&](Event) -> bool {
    return false;
  });

  Loop loop(&screen, component);

  while (!loop.HasQuitted()) {
    loop.RunOnce();
    screen.Post(Event::Custom);
    std::this_thread::sleep_for(std::chrono::milliseconds(10));
  }

  return EXIT_SUCCESS;
}
